# API Completeness

Narwhals has two different level of support for libraries: "full" and "interchange".

Libraries for which we have full support we intend to support the whole Narwhals API, however this is a work in progress.

In the following table it is possible to check which method is implemented for which backend. 

!!! info

    - "pandas-like" means pandas, cuDF and Modin
    - Polars supports all the methods (by design)

{{ backend_table }}
